#include "Pokitto.h"
#include "adpcm.h"
#include <File>

#define audioBufferSize 512

#define TIMER_32_0_IRQn 18          // for Timer setup
#include "timer_11u6x.h"
#include "clock_11u6x.h"

unsigned char *audioBuffer = (unsigned char *) 0x20000000;
int bufferOffset[4]={ audioBufferSize*3,  0, audioBufferSize, audioBufferSize*2 };

uint8_t myCurrentBuffer = 0;
uint8_t myCompleteBuffer = 0;         // Which section is full
long int audioOffset=0;

// messing with sound
int LISTENINGSAMPLERATE = 11025;   // samplerate of background music
bool playingMusic = false;
int globalVolume = 16;

File musicFile;    // Handle of open audio file
File dumpFile;

// WAVE file header format
struct HEADER {
    unsigned char riff[4];                      // RIFF string
    unsigned int overall_size   ;               // overall size of file in bytes
    unsigned char wave[4];                      // WAVE string
    unsigned char fmt_chunk_marker[4];          // fmt string with trailing null char
    unsigned int length_of_fmt;                 // length of the format data
    unsigned int format_type;                   // format type. 1-PCM, 3- IEEE float, 6 - 8bit A law, 7 - 8bit mu law
    unsigned int channels;                      // no.of channels
    unsigned int sample_rate;                   // sampling rate (blocks per second)
    unsigned int byterate;                      // SampleRate * NumChannels * BitsPerSample/8
    unsigned int block_align;                   // NumChannels * BitsPerSample/8
    unsigned int bits_per_sample;               // bits per sample, 8- 8bits, 16- 16 bits etc
    unsigned char data_chunk_header [4];        // DATA string or FLLR string
    unsigned int data_size;                     // NumSamples * NumChannels * BitsPerSample/8 - size of the next chunk that will be read
};
struct HEADER header;


// enableDAC() from Pokitto MiniLib
inline void enableDAC() {
    volatile unsigned int *PIO1 = (volatile unsigned int *) 0x40044060;
    volatile unsigned int *PIO2 = (volatile unsigned int *) 0x400440F0;
    volatile unsigned int *DIR1 = (volatile unsigned int *) 0xA0002004;
    volatile unsigned int *DIR2 = (volatile unsigned int *) 0xA0002008;
    PIO1[28] = PIO1[29] = PIO1[30] = PIO1[31] = 1<<7;
    PIO2[20] = PIO2[21] = PIO2[22] = PIO2[23] = 1<<7;
    *DIR1 |= (1<<28) | (1<<29) | (1<<30) | (1<<31);
    *DIR2 |= (1<<20) | (1<<21) | (1<<22) | (1<<23);

    // DIR1 is already declared above
    volatile unsigned int* SET1 = (unsigned int*)(0xa0002204);
    *DIR1 |= 1 << 17;
    *SET1 = 1 << 17;
    
}

// writeDAC() from Pokitto MiniLib
inline void writeDAC(unsigned char out) {
    volatile unsigned char* P1 = (unsigned char*)(0xa0000020);
    volatile unsigned char* P2 = (unsigned char*)(0xa0000040);
    P1[28] = out & 1; out >>= 1;
    P1[29] = out & 1; out >>= 1;
    P1[30] = out & 1; out >>= 1;
    P1[31] = out & 1; out >>= 1;
    P2[20] = out & 1; out >>= 1;
    P2[21] = out & 1; out >>= 1;
    P2[22] = out & 1; out >>= 1;
    P2[23] = out;
}

void clearAudioBuffer(){
    for(int t=0; t<audioBufferSize*4; t++){
        audioBuffer[t] = 127;
    }
}

    uint8_t tempBuffer[audioBufferSize];
    ImaState state;
    
    int dataChunckCounter = 0;
    unsigned char buffer4[4];


void audioTimer(void){
	if (Chip_TIMER_MatchPending(LPC_TIMER32_0, 1)) {
        
        writeDAC( audioBuffer[audioOffset] );
        
        myCurrentBuffer = audioOffset/audioBufferSize;
        
        if( myCompleteBuffer != myCurrentBuffer){
            myCompleteBuffer = myCurrentBuffer;

            musicFile.read(tempBuffer, sizeof(tempBuffer));
            int t=0;
            for(int s=0; s<sizeof(tempBuffer); s++){
                // Pokitto uses 8bit unsigned audio, the output is 16bit signed
                int number = ImaAdpcmDecode(tempBuffer[s]&15, state);
                audioBuffer[bufferOffset[myCompleteBuffer]+t++] = (number/256)+128;
                number = ImaAdpcmDecode(tempBuffer[s]>>4, state);
                audioBuffer[bufferOffset[myCompleteBuffer]+t++] = (number/256)+128;

                // if we'vre read a whole data block, read the next one
                if(++dataChunckCounter == header.block_align -4 ){
                    dataChunckCounter = 0;
                    musicFile.read(buffer4, sizeof(buffer4));
                    state.previousValue = (buffer4[0]) | (buffer4[1] << 8); // current sample
                    state.index = buffer4[2]; // step size table ref
                }
            }
            dumpFile.write(&audioBuffer[bufferOffset[myCompleteBuffer]],audioBufferSize);

        }
        
        
        if(++audioOffset == audioBufferSize*4){
            audioOffset = 0;
        }
        
        
        
 	    Chip_TIMER_ClearMatch(LPC_TIMER32_0, 1);
    }
}

// timer init stolen directly from Pokittolib
void initTimer(uint32_t sampleRate){
    /* Initialize 32-bit timer 0 clock */
	Chip_TIMER_Init(LPC_TIMER32_0);
    /* Timer rate is system clock rate */
	uint32_t timerFreq = Chip_Clock_GetSystemClockRate();
	/* Timer setup for match and interrupt at TICKRATE_HZ */
	Chip_TIMER_Reset(LPC_TIMER32_0);
	/* Enable both timers to generate interrupts when time matches */
	Chip_TIMER_MatchEnableInt(LPC_TIMER32_0, 1);
    /* Setup 32-bit timer's duration (32-bit match time) */
	Chip_TIMER_SetMatch(LPC_TIMER32_0, 1, (timerFreq / sampleRate));
	/* Setup both timers to restart when match occurs */
	Chip_TIMER_ResetOnMatchEnable(LPC_TIMER32_0, 1);
	/* Start both timers */
	Chip_TIMER_Enable(LPC_TIMER32_0);
	/* Clear both timers of any pending interrupts */
	NVIC_ClearPendingIRQ((IRQn_Type)TIMER_32_0_IRQn);
    /* Redirect IRQ vector - Jonne*/
    NVIC_SetVector((IRQn_Type)TIMER_32_0_IRQn, (uint32_t)&audioTimer);
	/* Enable both timer interrupts */
	NVIC_EnableIRQ((IRQn_Type)TIMER_32_0_IRQn);
}



void readWAVHeader(const char* fileName){
    
    unsigned char buffer4[4];
    unsigned char buffer2[2];

    printf("Opening  file..\n");

    if(musicFile.openRO(fileName)){
        printf(fileName);
        printf(" is open\n");
        
        musicFile.read(header.riff, sizeof(header.riff));
        printf("(1-4): %s \n", header.riff);

        musicFile.read(buffer4, sizeof(buffer4));
        printf("%u %u %u %u\n", buffer4[0], buffer4[1], buffer4[2], buffer4[3]);

        // convert little endian to big endian 4 byte int
        header.overall_size  = buffer4[0] | (buffer4[1]<<8) | (buffer4[2]<<16) | (buffer4[3]<<24);

        printf("(5-8) Overall size: bytes:%u, Kb:%u \n", header.overall_size, header.overall_size/1024);

        musicFile.read(header.wave, sizeof(header.wave));
        printf("(9-12) Wave marker: %s\n", header.wave);

        musicFile.read(header.fmt_chunk_marker, sizeof(header.fmt_chunk_marker));
        printf("(13-16) Fmt marker: %s\n", header.fmt_chunk_marker);
        
        musicFile.read(buffer4, sizeof(buffer4));
        printf("%u %u %u %u\n", buffer4[0], buffer4[1], buffer4[2], buffer4[3]);

        // convert little endian to big endian 4 byte integer
        header.length_of_fmt = buffer4[0] | (buffer4[1] << 8) | (buffer4[2] << 16) | (buffer4[3] << 24);
        printf("(17-20) Length of Fmt header: %u \n", header.length_of_fmt);
        // current location (20) + header.length_of_fmt = locations of 'fact' data


        musicFile.read(buffer2, sizeof(buffer2));
        printf("%u %u \n", buffer2[0], buffer2[1]);

        header.format_type = buffer2[0] | (buffer2[1] << 8);
        char format_name[10] = "";
        if (header.format_type == 1)
            strcpy(format_name,"PCM");
        else if (header.format_type == 2)
            strcpy(format_name, "MS ADPCM");
        else if (header.format_type == 6)
            strcpy(format_name, "A-law");
        else if (header.format_type == 7)
            strcpy(format_name, "Mu-law");
        else if (header.format_type == 17)
            strcpy(format_name, "IMA ADPCM");

        printf("(21-22) Format type: %u %s \n", header.format_type, format_name);

        musicFile.read(buffer2, sizeof(buffer2));
        printf("%u %u \n", buffer2[0], buffer2[1]);

        header.channels = buffer2[0] | (buffer2[1] << 8);
        printf("(23-24) Channels: %u \n", header.channels);

        musicFile.read(buffer4, sizeof(buffer4));
        printf("%u %u %u %u\n", buffer4[0], buffer4[1], buffer4[2], buffer4[3]);

        header.sample_rate = buffer4[0] | (buffer4[1] << 8) | (buffer4[2] << 16) | (buffer4[3] << 24);

        printf("(25-28) Sample rate: %u\n", header.sample_rate);

        musicFile.read(buffer4, sizeof(buffer4));
        printf("%u %u %u %u\n", buffer4[0], buffer4[1], buffer4[2], buffer4[3]);

        header.byterate  = buffer4[0] | (buffer4[1] << 8) | (buffer4[2] << 16) |  (buffer4[3] << 24);
        printf("(29-32) Byte Rate: %u , Bit Rate:%u\n", header.byterate, header.byterate*8);

        musicFile.read(buffer2, sizeof(buffer2));
        printf("%u %u \n", buffer2[0], buffer2[1]);

        header.block_align = buffer2[0] | (buffer2[1] << 8);
        printf("(33-34) Block Alignment: %u \n", header.block_align);
        
        musicFile.read(buffer2, sizeof(buffer2));
        printf("%u %u \n", buffer2[0], buffer2[1]);

        header.bits_per_sample = buffer2[0] | (buffer2[1] << 8);
        printf("(35-36) Bits per sample: %u \n", header.bits_per_sample);
        
        
        if(header.format_type != 1){ // PCM
            // if not PCM //

            if(header.format_type == 17){ // MS ADPCM

                // extra data
                // 2 bytes - bits per sample - should be 4 for ADPCM
                musicFile.read(buffer2, sizeof(buffer2));
                printf("Size of Extra Data %d \n", buffer2[0] | (buffer2[1] << 8));
    
                // 2 bytes - samples per block
                musicFile.read(buffer2, sizeof(buffer2));
                printf("Samples per Block %d \n", buffer2[0] | (buffer2[1] << 8));
    
                // 2 bytes - number of coefficient sets used in encoding
            //    musicFile.read(buffer2, sizeof(buffer2));
            //    int coeff = buffer2[0] | (buffer2[1] << 8);
            //    printf("sets used %d \n", coeff);
            
                // I don't know what this data is for, it's not used by the decoder
            /*
                for(int i=0; i<coeff; i++){
                    musicFile.read(buffer2, sizeof(buffer2));
                    short signed int number = buffer2[0] | (buffer2[1] << 8);
                    printf("set %d : %d, ", i, number);
                    
                    musicFile.read(buffer2, sizeof(buffer2));
                    number = buffer2[0] | (buffer2[1] << 8);
                    printf("%d, \n", number);
                }
                */
                // 4 bytes - 'fact' - sub chunk 2 id
                musicFile.read(buffer4, sizeof(buffer4));
                printf("Fact? %s \n", buffer4);
                // 4 bytes - sub chunk 2 size
                unsigned int number;
                musicFile.read(buffer4, sizeof(buffer4));
                number = buffer4[0] | (buffer4[1] << 8) | (buffer4[2] << 16) |  (buffer4[3] << 24);
                printf("last data size %d \n", number);
    
                for(int i=0; i<number; i++){
                    uint8_t temp1;
                    musicFile.read(&temp1, sizeof(temp1));
                }
            }        
        
        // 4 bytes - 'data'

        }

        
        musicFile.read(header.data_chunk_header, sizeof(header.data_chunk_header));
        printf("(37-40) Data Marker: %s \n", header.data_chunk_header);
        
        musicFile.read(buffer4, sizeof(buffer4));
        printf("%u %u %u %u\n", buffer4[0], buffer4[1], buffer4[2], buffer4[3]);
        
        header.data_size = buffer4[0] | (buffer4[1] << 8) | (buffer4[2] << 16) | (buffer4[3] << 24 );
        printf("(41-44) Size of data chunk: %u \n", header.data_size);

        // calculate no.of samples
        long num_samples = (8 * header.data_size) / (header.channels * header.bits_per_sample);
        printf("Number of samples:%lu \n", num_samples);
        
        long size_of_each_sample = (header.channels * header.bits_per_sample) / 8;
        printf("Size of each sample:%ld bytes\n", size_of_each_sample);
        

        if(header.format_type == 1){ // PCM
            // do 'Normal' uncompressed audio playing here...
        }

        if(header.format_type == 2){ // MS ADPCM
            initTimer(header.sample_rate);
        }
        if(header.format_type == 17){ // IMA ADPCM
            initTimer(header.sample_rate);
        }

    }

}



int main(){
    using PC=Pokitto::Core;
    using PD=Pokitto::Display;
    PC::begin();
    PD::persistence = true;
    PD::invisiblecolor = 0;

    clearAudioBuffer();

    readWAVHeader("imaadpcm.wav");

    dumpFile.openRW("dump.raw",1,1);
    
    

    // read first 4 bytes
    /* Packet Header (Mono)
	Byte	Description
	1	LSB of current sample (always zero in Microsoft version
	1	MSB of current sample
	1	Current step size
	1	Always Zero
	n	Compressed samples
    */
    musicFile.read(buffer4, sizeof(buffer4));

    state.previousValue = 0;//(buffer4[0]) | (buffer4[1] << 8); // current sample
    state.index = 0;//buffer4[2]; // step size table ref

    while( PC::isRunning() ){
        if( !PC::update() ) 
            continue;


    }
    
    return 0;
}
